Reference :
https://github.com/intel/asynch_mode_nginx#readme


Asynch Mode for NGINX-QUIC\*
## Table of Contents
- [Introduction](#introduction)
- [Licensing](#licensing)
- [Features](#features)
- [Hardware Requirements](#hardware-requirements)
- [Software Requirements](#software-requirements)
- [Additional Information](#additional-information)
- [Limitations](#limitations)
- [Installation Instructions](#installation-instructions)
    - [Asynch Mode for NGINX-QUIC\* Installation](#asynch-mode-for-nginx-quic-installation)
    - [Build BoringSSL\* and QAT engine](#build-boringssl-and-qat-engine)
- [SSL Engine Framework for Configuration](#ssl-engine-framework-for-configuration)
- [QAT Sample Configuration](#qat-sample-configuration)
- [Known Issues](#known-issues)
- [Intended Audience](#intended-audience)
- [Legal](#legal)

## Introduction
NGINX-QUIC\* is a preview of QUIC+HTTP/3 for NGINX\*, one popular web server originally written by Igor Sysoev.
Similar to the NGINX mainline OpenSSL\* extension at https://github.com/intel/asynch_mode_nginx, an asynchronous mode BoringSSL\* extension accelerated by Intel QuickAssist Technology (QAT) is also provided, and it shows significant handshake performance improvement.

## Licensing
The Licensing of the files within this project is:

Asynch Mode for NGINX-QUIC* - BSD License. Please see the `LICENSE` file contained in the top level folder. Further details can be found in the file headers of the relevant files.

## Features
* Asynchronous Mode in SSL/TLS/QUIC processing (including http/proxy module)
* SSL Engine Framework for engine configuration, and QAT module for detailed device setup
* Compatible asynchronous mode directives between this NGINX-QUIC and [asynch_mode_nginx][1]
* Support software fallback for asymmetric cryptography algorithms.

[1]:https://github.com/intel/asynch_mode_nginx

## Hardware Requirements
Asynch Mode for NGINX-QUIC* supports Crypto offload to the following acceleration devices:
* [Intel&reg; C62X Series Chipset][2]

[2]:https://www.intel.com/content/www/us/en/design/products-and-solutions/processors-and-chipsets/purley/intel-xeon-scalable-processors.html

## Software Requirements
This release was validated on the following:
* Asynch Mode for NGINX-QUIC* has been tested with Intel® QuickAssist Acceleration Driver QAT1.7.L.4.14.0-00031. Please download the QAT driver from the link https://01.org/intel-quickassist-technology
* BoringSSL: commit 5e7229488844e987207b377968b3cf0340bc4ccf

## Additional Information
* Asynch Mode for NGINX-QUIC* is developed based on the Intel® QuickAssist Technology BoringSSL* Library.

* Asynch Mode for NGINX-QUIC* SSL engine framework provides new directives:

**Directives**
```bash
Syntax:     ssl_asynch on | off;
Default:    ssl_asynch off;
Context:    http, server
Enables SSL/TLS asynchronous mode
```
**Example**

file: conf/nginx.conf
```bash
    http {
        ssl_asynch  on;
        server {...}
    }
```

**Directives**
```bash
Syntax:     proxy_ssl_asynch on | off;
Default:    proxy_ssl_asynch off;
Context:   http, server
Enables the SSL/TLS protocol asynchronous mode for connections to a proxied server.
```

**Example**

file: conf/nginx.conf
```bash
    http {
        server {
            location /ssl {
                proxy_pass https://127.0.0.1:8082/outer;
                proxy_ssl_asynch on;
            }
        }
    }
```

* Asynch Mode for NGINX-QUIC\* provide new option `asynch` for `listen` directive.

**Example**

file: conf/nginx.conf
```bash
    http {
        server{ listen 443 asynch; }
    }
```

## Limitations
* Nginx supports `reload` operation, when QAT hardware is involved for crypto offload, user should ensure that there are enough number of QAT instances. For example, the available qat instance number should be 2x equal or more than Nginx worker process number.

  For example, in Nginx configuration file (`nginx.conf`) worker process number is configured as
   ```bash
   worker_processes 16;
   ```
  Then the instance configuration in QAT driver configuration file should be
    ```bash
    [SHIM]
    NumberCyInstances = 1
    NumberDcInstances = 0
    NumProcesses = 32
    LimitDevAccess = 0
    ```
* When configure "worker_process auto", Asynch Mode for NGINX-QUIC\* will need instance number equal or larger than 2 times of CPU core number. Otherwise, Asynch Mode for NGINX-QUIC\* might show various issues caused by leak of instance.
* The directive `asynch` is not compatible with the directive `http3`.

## Installation Instructions

### Asynch Mode for NGINX-QUIC* Installation

**Set the following environmental variables:**

`NGINX_INSTALL_DIR` is the directory where Nginx will be installed to

`BORINGSSL_LIB` is the directory where the BoringSSL has been installed to

`BORINGSSL_SRC` is the directory where the BoringSSL has been cloned to

`QAT_ENGINE_LIB` is the directory where the QATEngine library has been installed to

`QAT_ENGINE_SRC` is the source code directory where the QATEngine has been downloaded to

**Configure nginx for compilation:**
```bash
    ./configure \
        --prefix=$NGINX_INSTALL_DIR \
        --with-http_v3_module \
        --add-dynamic-module=modules/ngx_bssl_qat_module \
        --with-cc-opt="-I$BORINGSSL_SRC/include -I$QAT_ENGINE_SRC" \
        --with-ld-opt="-Wl,-rpath=$BORINGSSL_LIB -L$BORINGSSL_LIB  \
                       -Wl,-rpath=$QAT_ENGINE_LIB -L$QAT_ENGINE_LIB"
```

**Compile and Install:**

```bash
    make
    make install
```

**Nginx supports setting worker to non-root user, for example:**

    Add user qat in group qat, for example run below command in your terminal:
    ```bash
    groupadd qat
    useradd -g qat qat
    ```
    In nginx.conf, you can set worker as qat, qat is the user you added before:
    ```bash
    user qat qat;
    ```
    Then we need to give non-root worker enough permission to enable qat, you need to run follow commands in your terminal:
    ```bash
    chgrp qat /dev/qat_*
    chmod 660 /dev/qat_*
    chgrp qat /dev/usdm_drv
    chmod 660 /dev/usdm_drv
    chgrp qat /dev/uio*
    chmod 660 /dev/uio*
    chgrp qat /usr/local/lib/libqat_s.so
    chgrp qat /usr/local/lib/libusdm_drv_s.so
    ```
### Build BoringSSL* and QAT engine
These instructions can be found in Intel® QuickAssist Technology(QAT) BoringSSL* Library.

## SSL Engine Framework for Configuration
An SSL Engine Framework is introduced to provide a powerful and flexible mechanism to configure Nginx SSL engine directly in the Nginx configuration file (`nginx.conf`).

While BoringSSL* is enabled, QAT engine is a set of private key methods to enable by applications, then the overall configuration comes from Nginx file (`nginx.conf`).

### ssl_engine configuration
A new configuration block is introduced as `ssl_engine` which provides two general the directive:

Sets the engine module and engine id for SSL async engine by only providing the engine id.
```bash
Syntax:     use_engine [engine id];
Default:    N/A
Context:    ssl_engine
Description:
            Specify the engine id
```

```bash
Syntax:     default_algorithms [RSA,EC];
Default:    RSA
Context:    ssl_engine
Description:
            Specify the algorithms need to be offloaded to the engine
```

A following configuration sub-block can be used to set engine specific configuration.

### ngx_bssl_qat_module
Any 3rd party modules can be integrated into this framework. By default, a QAT module `ngx_bssl_qat_module` is provided in `modules/ngx_bssl_qat_module`.

An example configuration in the `nginx.conf`:
```bash
  load_module modules/ngx_bssl_qat_module.so;
    ...
    ssl_engine {
        use_engine qatengine;
        default_algorithms RSA,EC;
        qat_engine {
            qat_sw_fallback on;
            qat_offload_mode async;
            qat_notify_mode poll;
            qat_poll_mode internal;
        }
    }
```

## QAT Sample Configuration
file: `conf/nginx-cpu-sync.conf`

This is a sample configure file shows how to configure QAT in nginx.conf. This file includes:
* Disable SSL QAT engine
* Select TLS-1.3 as the default ssl_protocols.

file: `conf/nginx-qat-async.conf`

This is a sample configure file shows how to configure QAT in nginx.conf. This file includes:
* Enable SSL QAT engine in internal mode.
* Support HTTPS/QUIC async mode.
* Select TLS-1.3 as the default ssl_protocols.

file: `conf/nginx-qat-sync.conf`

This is a sample configure file shows how to configure QAT in nginx.conf. This file includes:
* Enable SSL QAT engine in internal mode (inline mode is available).
* Support HTTPS/QUIC sync mode.
* Select TLS-1.3 as the default ssl_protocols.

## Known Issues
**Asynchronous Mode SSL/TLS is not enabled for all services (stream, mail) yet**<br/>

## Intended Audience
The target audience may be software developers, test and validation engineers, system integrators, end users and consumers for Asynch Mode for NGINX-QUIC* integrated Intel® Quick Assist Technology.

## Legal
Intel disclaims all express and implied warranties, including without limitation, the implied warranties of merchantability, fitness for a particular purpose, and non-infringement, as well as any warranty arising from course of performance, course of dealing, or usage in trade.

This document contains information on products, services and/or processes in development. All information provided here is subject to change without notice. Contact your Intel representative to obtain the latest forecast , schedule, specifications and roadmaps.

The products and services described may contain defects or errors known as errata which may cause deviations from published specifications. Current characterized errata are available on request.

Copies of documents which have an order number and are referenced in this document may be obtained by calling 1-800-548-4725 or by visiting www.intel.com/design/literature.htm.

Intel, the Intel logo are trademarks of Intel Corporation in the U.S. and/or other countries.

\*Other names and brands may be claimed as the property of others
